package htsjdk.beta.plugin.reads;

/**
 * Class with string constants for known formats supported by reads codecs.
 */
public class ReadsFormats {

    /**
     * SAM file format.
     */
    public static final String SAM = "SAM";

    /**
     * BAM file format.
     */
    public static final String BAM = "BAM";

    /**
     * CRAM file format.
     */
    public static final String CRAM = "CRAM";

    /**
     * GA4GH htsget BAM format.
     */
    public static final String HTSGET_BAM = "HTSGET_BAM";

}
