package htsjdk.beta.plugin.hapref;

import htsjdk.beta.plugin.HtsEncoderOptions;

/**
 * Class for haploid reference encoder options.
 */
public class HaploidReferenceEncoderOptions implements HtsEncoderOptions {
}
