package htsjdk.beta.plugin;

//Note: This is a just a tagging interface used to tag SAMFileHeader, VCFHeader, and SAMSequenceDictionary.
// It will be more useful once version upgrading is implemented.

/**
 * Tagging interface used as a type-bound for codec/encoder/decoder header type params.
 */
public interface HtsHeader {
}
