package htsjdk.beta.plugin.variants;

/**
 * Class with string constants representing known formats supported by variants codecs.
 */
public class VariantsFormats {
    /**
     * VCF format.
     */
    public static final String VCF = "VCF";

    /**
     * BCF format.
     */
    public static final String BCF = "BCF";
}
