package htsjdk.beta.plugin.hapref;

/**
 * Class with string constants representing the file formats supported by haploid reference codecs.
 */
public class HaploidReferenceFormats {

    /**
     * Fasta format
     */
    public static final String FASTA = "FASTA";

}
