/*
 * The MIT License
 *
 * Copyright (c) 2009 The Broad Institute
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package htsjdk.samtools.util;

import java.io.Closeable;

/**
 * Interface allows for implementations that read lines from a String, an ASCII file, or somewhere else.
 */
public interface LineReader extends Closeable {

    // value to return in call to peek, if eof has been reached.
    int EOF_VALUE = -1;

    /**
     * Read a line and remove the line terminator
     * @return the line read, or null if EOF has been reached.
     */
    String readLine();

    /**
     * @return 1-based number of line most recently read
     */
    int getLineNumber();

    /**
     * Non-destructive one-character look-ahead.
     * @return If not eof, the next character that would be read.  If eof, {@value EOF_VALUE}.
     */
    int peek();

    @Override
    public void close();
}
