package htsjdk.beta.plugin.registry.testcodec;

// File formats for the HtsTest codec implementations used to test codec resolution
public class HtsTestCodecFormats {

    public final static String FILE_FORMAT_1 = "FILE_FORMAT_1";
    public final static String FILE_FORMAT_2 = "FILE_FORMAT_2";
    public final static String FILE_FORMAT_3 = "FILE_FORMAT_3";

}
