package htsjdk.beta.plugin.hapref;

import htsjdk.beta.plugin.HtsDecoderOptions;

/**
 * Class for haploid reference decoder options.
 */
public class HapRefDecoderOptions implements HtsDecoderOptions {
}
